/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * This file is part of SableCC.                             *
 * See the file "LICENSE" for copyright information and the  *
 * terms and conditions for copying, distribution and        *
 * modification of SableCC.                                  *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

package org.sablecc.sablecc;

import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.Map;

import org.sablecc.sablecc.analysis.DepthFirstAdapter;
import org.sablecc.sablecc.node.*;

public class GenerateAlternativeCodeForParser extends DepthFirstAdapter
{
  String currentAlt;
  String realcurrentAlt;
  BufferedWriter file;
  private File pkgDir;

  private ResolveTransformIds transformIds;
  private ComputeCGNomenclature CG;
  private ComputeSimpleTermPosition CTP;
  private MacroExpander macros;
  private Map<Node, String> simpleTermTransformMap;
  private List<Node> listSimpleTermTransform;
  private Map<Node, String> simpleTermOrsimpleListTermTypes;

  private int popCount;

  GenerateAlternativeCodeForParser(File pkgDir, String aParsedAltName,
                                   String raParsedAltName,
                                   BufferedWriter file,
                                   ResolveTransformIds transformIds,
                                   ComputeCGNomenclature CG,
                                   ComputeSimpleTermPosition CTP,
                                   Map<Node, String> simpleTermTransformMap,
                                   MacroExpander macros,
                                   List<Node> listSimpleTermTransform,
                                   Map<Node, String> simpleTermOrsimpleListTermTypes)
  {
    this.pkgDir = pkgDir;
    this.file = file;
    currentAlt = aParsedAltName;
    realcurrentAlt = raParsedAltName;
    this.transformIds = transformIds;
    this.CG = CG;
    this.CTP = CTP;
    this.simpleTermTransformMap = simpleTermTransformMap;
    this.macros = macros;
    this.listSimpleTermTransform = listSimpleTermTransform;
    this.simpleTermOrsimpleListTermTypes = simpleTermOrsimpleListTermTypes;

    this.popCount = 0;
  }

  @Override
  public void caseAElem(final AElem node) {
    popCount++;
  }

  private String lookupInternalTypeName(Node term) {
    String name = simpleTermTransformMap.get(term);
    if(name != null)
    {
      return name;
    }
    else
    {
      return CG.getAltTransformElemTypes().get(term);
    }
  }

  private static String getVariableName(String typeName, int position)
  {
    if("null".equals(typeName))
    {
      return "null";
    }
    else if("Lnull".equals(typeName))
    {
      return "Collections.emptyList()";
    }
    else if(typeName.startsWith("L"))
    {
      return "listNode" + position;
    }
    else
    {
      return typeName.substring(0, 1).toLowerCase() + typeName.substring(1) + "Node" + position;
    }
  }

  private String getVariableName(Node term)
  {
    return getVariableName(lookupInternalTypeName(term), CG.getTermNumbers().get(term));
  }

  private static String variableTypeFromInternalType(String typeName)
  {
    if("null".equals(typeName))
    {
      throw new IllegalArgumentException("Null values have no variable type");
    }
    else if(typeName.startsWith("L"))
    {
      String elementTypeName = typeName.substring(1);
      if("null".equals(elementTypeName))
      {
        throw new IllegalArgumentException("Empty list values have no variable type");
      }
      else
      {
        return "List<" + elementTypeName + ">";
      }
    }
    else
    {
      return typeName;
    }
  }

  @Override
  public void outAAltTransform(AAltTransform node)
  {
    try
    {
      if(node.getTerms().isEmpty())
      {
        macros.apply(file, "ParserNewTailEmpty", null);
      }
      else
      {
        PTerm firstTerm = node.getTerms().get(0);
        if (popCount > 0) {
          // The nodeArrayList variables are numbered starting at 1, so the first popped variable has the number popCount and not popCount-1.
          macros.apply(file, "ParserNewComputePositions", new String[] {getVariableName(firstTerm), String.valueOf(popCount)});
        }

        if(node.getTerms().size() == 1)
        {
          macros.apply(file, "ParserNewTailSingle", new String[] {getVariableName(firstTerm)});
        }
        else
        {
          macros.apply(file, "ParserNewTailMultiHead");
          for(PTerm term : node.getTerms())
          {
            macros.apply(file, "ParserNewTailMultiElement", new String[] {getVariableName(term)});
          }
          macros.apply(file, "ParserNewTailMultiTail");
        }
      }
    }
    catch(IOException e)
    {
      throw new RuntimeException("An error occured while writing to " +
                                 new File(pkgDir, "Parser.java").getAbsolutePath(), e);
    }
  }

  private void generateSimpleTerm(Node node, TId id, TId simpleTermTail)
  {
    try
    {
      String type_name = lookupInternalTypeName(node);
      String termKey = currentAlt + "." + id.getText();
      int elemPosition = CTP.elems_position.get(termKey);
      int positionMap = 0;

      if(simpleTermTail != null)
      {
        // FIXME This condition is always false, because listSimpleTermTransform contains ASimpleTerm/ASimpleListTerm and not String! Is it safe to remove this?
        if(!listSimpleTermTransform.contains(id.getText()))
        {
          String type = CTP.positionsMap.get(realcurrentAlt + "." + id.getText());
          List<String> list = transformIds.getProdTransformIds().prod_transforms.get(type);
          if(list.indexOf(simpleTermTail.getText()) >= 0)
          {
            positionMap = list.indexOf(simpleTermTail.getText());
          }
        }

        if(simpleTermOrsimpleListTermTypes.get(node) != null)
        {
          String type = simpleTermOrsimpleListTermTypes.get(node);
          List<String> list = transformIds.getProdTransformIds().prod_transforms.get(type);
          if(list.indexOf(simpleTermTail.getText()) >= 0)
          {
            positionMap = list.indexOf(simpleTermTail.getText());
          }
        }
      }

      if("null".equals(type_name))
      {
        throw new RuntimeException("Simple term has null type - this should never happen");
      }
      else if("Lnull".equals(type_name))
      {
        throw new RuntimeException("Simple term has list type with no element type - this should never happen");
      }

      macros.apply(file, "ParserSimpleTerm", new String[]
                   {
                     variableTypeFromInternalType(type_name),
                     getVariableName(node),
                     ""+elemPosition, ""+positionMap
                   }
                  );
    }
    catch(IOException e)
    {
      throw new RuntimeException("An error occured while writing to " +
                                 new File(pkgDir, "Parser.java").getAbsolutePath(), e);
    }
  }

  @Override
  public void inASimpleTerm(ASimpleTerm node)
  {
    generateSimpleTerm(node, node.getId(), node.getSimpleTermTail());
  }

  @Override
  public void inASimpleListTerm(ASimpleListTerm node)
  {
    generateSimpleTerm(node, node.getId(), node.getSimpleTermTail());
  }

  @Override
  public void caseAListTerm(AListTerm node)
  {
    if(node.getListTerms().isEmpty())
    {
      // Empty lists are generated inline and don't have a variable.
      return;
    }

    try
    {
      for(PListTerm listTerm : node.getListTerms())
      {
        listTerm.apply(this);
      }

      int nonNullTermsCount = 0;
      PListTerm firstNonNullListTerm = null;
      String firstNonNullTypeName = null;
      for(PListTerm listTerm : node.getListTerms())
      {
        String typeName = lookupInternalTypeName(listTerm);
        // Null and empty list terms never add any elements to the list and so don't need to be counted.
        if("null".equals(typeName) || "Lnull".equals(typeName))
        {
          continue;
        }
        nonNullTermsCount++;
        if(firstNonNullListTerm == null)
        {
          firstNonNullListTerm = listTerm;
          firstNonNullTypeName = typeName;
        }
      }

      if(nonNullTermsCount == 0)
      {
        throw new RuntimeException("List has non-empty type, but contains no elements!");
      }
      assert firstNonNullListTerm != null;
      assert firstNonNullTypeName != null;

      String destType = lookupInternalTypeName(node);

      if(nonNullTermsCount == 1)
      {
        if(firstNonNullTypeName.startsWith("L"))
        {
          macros.apply(file, "ParserTypedListAssign", new String[] {
            variableTypeFromInternalType(destType), getVariableName(node), getVariableName(firstNonNullListTerm),
          });
        }
        else
        {
          macros.apply(file, "ParserTypedSingleElementList", new String[] {
            variableTypeFromInternalType(destType), getVariableName(node), getVariableName(firstNonNullListTerm),
          });
        }
      }
      else
      {
        boolean first = true;
        for(PListTerm listTerm : node.getListTerms())
        {
          String typeName = lookupInternalTypeName(listTerm);
          // Null and empty list terms never add any elements to the list.
          if("null".equals(typeName) || "Lnull".equals(typeName))
          {
            continue;
          }

          if(typeName.startsWith("L"))
          {
            if(first)
            {
              macros.apply(file, "ParserTypedLinkedListInitAddAll", new String[] {
                variableTypeFromInternalType(destType), getVariableName(node), getVariableName(listTerm),
              });
            }
            else
            {
              macros.apply(file, "ParserTypedLinkedListAddAll", new String[] {getVariableName(node), getVariableName(listTerm)});
            }
          }
          else
          {
            if(first)
            {
              macros.apply(file, "ParserTypedLinkedListInit", new String[] {
                variableTypeFromInternalType(destType), getVariableName(node),
              });
            }
            macros.apply(file, "ParserTypedLinkedListAdd", new String[] {getVariableName(node), getVariableName(listTerm)});
          }
          first = false;
        }
      }
    }
    catch(IOException e)
    {
      throw new RuntimeException("An error occured while writing to " +
                                 new File(pkgDir, "Parser.java").getAbsolutePath(), e);
    }
  }

  private void generateNewTerm(Node node, AProdName prodName, List<PTerm> params)
  {
    String newAltName = name(prodName);

    try
    {
      for(PTerm param : params)
      {
        param.apply(this);
      }

      macros.apply(file, "ParserNewBodyNew", new String[] {newAltName, getVariableName(node)});

      if(!params.isEmpty())
      {
        String isNotTheFirstParam = "";

        for(PTerm term : params)
        {
          macros.apply(file, "ParserNewBodyParam", new String[] {isNotTheFirstParam + getVariableName(term)});

          isNotTheFirstParam = ", ";
        }
      }
      macros.apply(file, "ParserNewBodyNewTail");
    }
    catch(IOException e)
    {
      throw new RuntimeException("An error occured while writing to " +
                                 new File(pkgDir, "TokenIndex.java").getAbsolutePath(), e);
    }
  }

  @Override
  public void caseANewTerm(ANewTerm node)
  {
    generateNewTerm(node, (AProdName)node.getProdName(), node.getParams());
  }

  @Override
  public void caseANewListTerm(ANewListTerm node)
  {
    generateNewTerm(node, (AProdName)node.getProdName(), node.getParams());
  }

  public String name(AProdName node)
  {
    if(node.getProdNameTail() != null)
    {
      return "A" +
             ResolveIds.name(node.getProdNameTail().getText()) +
             ResolveIds.name(node.getId().getText());
    }
    return "A" + ResolveIds.name(node.getId().getText());
  }

}
